import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin


class CustomEstimator(BaseEstimator, ClassifierMixin):

    def __init__(self, param1=0, param2=None):
        # it is important that you name your params the same in the constructor
        #  stub as well as in your class variables!
        self.param1 = param1
        self.param2 = param2

    def fit(self, X, y=None, **kwargs):
        """
        Adjust the underlying model or method to the data.

        Returns
        -------
        IMPORTANT: must return self!
        """
        return self

    def predict(self, X):
        """
        Use the learned model to make predictions.
        """
        return np.random.randint(0, 2, X.shape[0])



from typing import Callable
import warnings

from photonai.optimization import Categorical as PhotonCategorical
from photonai.optimization import FloatRange, IntegerRange, BooleanSwitch, PhotonHyperparam
from photonai.optimization.base_optimizer import PhotonMasterOptimizer
from photonai.photonlogger import logger

try:
    import nevergrad as ng
    from nevergrad.optimization.base import Optimizer
    __found__ = True
except ModuleNotFoundError:
    __found__ = False


class NevergradOptimizer(PhotonMasterOptimizer):
    """Nevergrad Wrapper for PHOTONAI.

    Nevergrad is a gradient-free optimization platform.

    Nevergrad [usage and implementation details](
    https://facebookresearch.github.io/nevergrad/).

    Example:
        ``` python
        import nevergrad as ng
        # list of all available nevergrad optimizer
        print(list(ng.optimizers.registry.values()))

        my_pipe = Hyperpipe('nevergrad_example',
                            optimizer='nevergrad',
                            optimizer_params={'facade': 'NGO', 'n_configurations': 30},
                            ...
                            )
        ```

    """
    def __init__(self, facade='NGO', n_configurations: int = 100, rng: int = 42):
        """
        Initialize the object.

        Parameters:
            facade:
                Choice of the Nevergrad backend strategy, e.g. [NGO, ...].

            n_configurations:
                Number of runs.

            rng:
                Random Seed.

        """

        if not __found__:
            msg = "Module nevergrad not found or not installed as expected. " \
                  "Please install the nevergrad/requirements.txt PHOTONAI provides."
            logger.error(msg)
            raise ModuleNotFoundError(msg)

        if facade in list(ng.optimizers.registry.values()):
            self.facade = facade
        elif facade in list(ng.optimizers.registry.keys()):
            self.facade = ng.optimizers.registry[facade]
        else:
            msg = "nevergrad.optimizer {} not known. Check out all available nevergrad optimizers " \
                  "by nevergrad.optimizers.registry.keys()".format(str(facade))
            logger.error(msg.format(str(facade)))
            raise ValueError(msg.format(str(facade)))

        self.n_configurations = n_configurations
        self.space = None  # Hyperparameter space for nevergrad
        self.switch_optiones = {}
        self.hyperparameters = []
        self.rng = rng

        self.maximize_metric = False
        self.constant_dictionary = {}

        self.objective = None
        self.optimizer = None

    def prepare(self, pipeline_elements: list, maximize_metric: bool, objective_function: Callable) -> None:
        """Prepare Nevergrad Optimizer.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

            objective_function:
                The cost or objective function.

        """
        self.space = self._build_nevergrad_space(pipeline_elements)
        self.space.random_state.seed(self.rng)
        if self.constant_dictionary:
            msg = "PHOTONAI has detected some one-valued params in your hyperparameters. Pleas use the kwargs for " \
                  "constant values. This run ignores following settings: " + str(self.constant_dictionary.keys())
            logger.warning(msg)
            warnings.warn(msg)
        self.maximize_metric = maximize_metric

        def nevergrad_objective_function(**current_config):
            return objective_function(current_config)
        self.objective = nevergrad_objective_function

        self.optimizer = self.facade(parametrization=self.space, budget=self.n_configurations)

    def optimize(self) -> None:
        self.optimizer.minimize(self.objective)

    def _build_nevergrad_space(self, pipeline_elements: list):
        """
        Build entire Nevergrad hyperparameter space.

        Parameters:
            pipeline_elements:
                List of all pipeline_elements to create the hyperparameter space.

        """
        param_dict = {}
        for pipe_element in pipeline_elements:
            # build conditions for switch elements
            if pipe_element.__class__.__name__ == 'Switch':
                raise NotImplementedError("Currently PHOTONAIs Switch is not supported by nevergrad.")

            if hasattr(pipe_element, 'hyperparameters'):
                for name, value in pipe_element.hyperparameters.items():
                    self.hyperparameters.append(name)
                    # if we only have one value we do not need to optimize
                    if isinstance(value, list) and len(value) < 2:
                        self.constant_dictionary[name] = value[0]
                        continue
                    if isinstance(value, PhotonCategorical) and len(value.values) < 2:
                        self.constant_dictionary[name] = value.values[0]
                        continue
                    nevergrad_param = self._convert_photonai_to_nevergrad_param(value)
                    if nevergrad_param is not None:
                        param_dict[name] = nevergrad_param
        return ng.p.Instrumentation(**param_dict)

    @staticmethod
    def _convert_photonai_to_nevergrad_param(hyperparam: PhotonHyperparam):
        """
        Helper function: Convert PHOTONAI to Nevergrad hyperparameter.

        Parameters:
            hyperparam:
                 One of photonai.optimization.hyperparameters.

        """
        if isinstance(hyperparam, PhotonCategorical) or isinstance(hyperparam, BooleanSwitch):
            return ng.p.Choice(hyperparam.values)
        elif isinstance(hyperparam, list):
            return ng.p.Choice(hyperparam)
        elif isinstance(hyperparam, FloatRange):
            if hyperparam.range_type == 'linspace':
                return ng.p.Scalar(lower=hyperparam.start, upper=hyperparam.stop)
            elif hyperparam.range_type == 'logspace':
                return ng.p.Log(lower=hyperparam.start, upper=hyperparam.stop)
            msg = str(hyperparam.range_type) + "in your float hyperparameter is not implemented yet."
            logger.error(msg)
            raise NotImplementedError(msg)
        elif isinstance(hyperparam, IntegerRange):
            return ng.p.Scalar(lower=hyperparam.start, upper=hyperparam.stop).set_integer_casting()

        msg = "Cannot convert hyperparameter " + str(hyperparam) + ". Supported types: Categorical, IntegerRange," \
                                                                   "FloatRange, list."
        logger.error(msg)
        raise ValueError(msg)



from itertools import product

from photonai.optimization import PhotonHyperparam, IntegerRange, FloatRange, Categorical, BooleanSwitch
from photonai.photonlogger import logger


def create_global_config_dict(pipeline_elements: list) -> dict:
    """
    Creation of a definition set for grid-based optimizers in format: dict key -> value.

    A grid is generated from a given list of hyperparameters for the optimization process.
    Furthermore, the initialization of hyperparameters takes place.

    Parameters:
        pipeline_elements:
            List of all set hyperparameters.

    Returns:
        Grid of configurations.

    """
    global_hyperparameter_dict = {}
    for p_element in pipeline_elements:

        # global_hyperparameter_dict[p_element.name] = {}

        if len(p_element.hyperparameters) > 0:
            for h_key, h_value in p_element.hyperparameters.items():
                if isinstance(h_value, list):
                    # global_hyperparameter_dict[p_element.name][h_key] = h_value
                    global_hyperparameter_dict[h_key] = h_value
                elif isinstance(h_value, PhotonHyperparam):
                    # when we have a range we need to convert it to a definite list of values
                    if isinstance(h_value, FloatRange) or isinstance(h_value, IntegerRange):
                        # build a definite list of values
                        h_value.transform()
                        # global_hyperparameter_dict[p_element.name][h_key] = h_value.values
                        global_hyperparameter_dict[h_key] = h_value.values
                    elif isinstance(h_value, BooleanSwitch) or isinstance(h_value, Categorical):
                        global_hyperparameter_dict[h_key] = h_value.values
    return global_hyperparameter_dict


def create_global_config_grid(pipeline_elements: list, add_name: str = '') -> list:
    """
    Creation of a list of configuration for grid-based optimizers.
    A grid is generated from a given list of hyperparameters for the optimization process.

    Parameters:
        pipeline_elements:
            List of PipelineElements.

        add_name: str, default=''
            Set prefix to dict keys.

    Returns:
        List of dicts. Every dict is a possible configurations.

    """
    global_hyperparameter_list = []
    for element in pipeline_elements:
        if hasattr(element, "generate_config_grid"):
            config_grid = element.generate_config_grid()
            if len(config_grid) > 0:
                global_hyperparameter_list.append(config_grid)

    praefix = ''
    if add_name != '':
        praefix = add_name + '__'

    # if len(global_hyperparameter_list) == 1:
    #     return [dict((praefix + pair[0], pair[1]) for d in global_hyperparameter_list[0] for pair in d.items())]
    # else:
    threshold = 1000000
    total_product_num = 1
    for i in global_hyperparameter_list:
        total_product_num = total_product_num * len(i)
    if total_product_num > threshold:
        msg = 'The entire configuration grid entails more than ' + str(threshold) + ' possible configurations. ' \
                                                                                    'This might take very ' \
                                                                                    'long to both compute ' \
                                                                                    'and process.'
        logger.error(msg)
        raise ValueError(msg)
    config_list = list(product(*global_hyperparameter_list))
    config_dicts = []
    # get all configs in one
    for c in config_list:
        config_dicts.append(dict((praefix + pair[0], pair[1]) for d in c for pair in d.items()))
    return config_dicts



from sklearn.datasets import load_boston
from sklearn.model_selection import KFold, ShuffleSplit

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange, Categorical

# WE USE THE BOSTON HOUSING DATA FROM SKLEARN
X, y = load_boston(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('skopt_example',
                    optimizer='sk_opt',  # which optimizer PHOTONAI shall use, in this case sk_opt
                    optimizer_params={'n_configurations': 25, 'acq_func': 'LCB', 'acq_func_kwargs': {'kappa': 1.96}},
                    metrics=['mean_squared_error', 'pearson_correlation'],
                    best_config_metric='mean_squared_error',
                    outer_cv=ShuffleSplit(n_splits=4, test_size=0.2),
                    inner_cv=KFold(n_splits=3),
                    verbosity=0,
                    project_folder='./tmp/')

# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe += PipelineElement('StandardScaler')

# engage and optimize SVR
# linspace and logspace is converted to uniform and log-uniform priors in skopt
my_pipe += PipelineElement('SVR', hyperparameters={'C': FloatRange(1e-3, 100, range_type='logspace'),
                                                   'epsilon': FloatRange(1e-3, 10, range_type='logspace'),
                                                   'tol': FloatRange(1e-4, 1e-2, range_type='linspace'),
                                                   'kernel': Categorical(['linear', 'rbf', 'poly'])})

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, OutputSettings
from photonai.optimization import FloatRange, Categorical, IntegerRange

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# YOU CAN SAVE THE TRAINING AND TEST RESULTS AND ALL THE PERFORMANCES IN THE MONGODB
mongo_settings = OutputSettings(mongodb_connect_url="mongodb://localhost:27017/photon_results")

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('basic_MongoDB_pipe',  # the name of your pipeline
                    optimizer='grid_search',
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=5),
                    verbosity=1,
                    project_folder='./tmp/',
                    # append the output_settings argument to hand over the mongo_setting to your pipeline
                    output_settings=mongo_settings)

# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe.add(PipelineElement('StandardScaler'))
# then do feature selection using a PCA, specify which values to try in the hyperparameter search
my_pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}, test_disabled=True)
# engage and optimize the good old SVM for Classification
my_pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                                   'C': FloatRange(0.5, 2, num=10)}, gamma='scale')

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



import warnings
from typing import Callable

from photonai.optimization import Categorical as PhotonCategorical
from photonai.optimization import FloatRange, IntegerRange, BooleanSwitch, PhotonHyperparam
from photonai.optimization.base_optimizer import PhotonMasterOptimizer
from photonai.photonlogger import logger

try:
    from smac.configspace import UniformFloatHyperparameter, UniformIntegerHyperparameter, CategoricalHyperparameter, \
        ConfigurationSpace, Configuration, InCondition, Constant
    from smac.scenario.scenario import Scenario
    from smac.facade.smac_bo_facade import SMAC4BO
    from smac.facade.smac_hpo_facade import SMAC4HPO
    from smac.facade.smac_ac_facade import SMAC4AC
    from smac.facade.smac_bohb_facade import BOHB4HPO
    __found__ = True
except (ModuleNotFoundError, ImportError):
    __found__ = False


class SMACOptimizer(PhotonMasterOptimizer):
    """SMAC Wrapper for PHOTONAI.

    SMAC (sequential model-based algorithm configuration) is a
    versatile tool for optimizing algorithm parameters.
    The main core consists of Bayesian Optimization in
    combination with an aggressive racing mechanism to efficiently
    decide which of two configurations performs better.

    SMAC usage and implementation details [here](
    https://automl.github.io/SMAC3/master/quickstart.html).

    References:

        Hutter, F. and Hoos, H. H. and Leyton-Brown, K.
        Sequential Model-Based Optimization for General Algorithm Configuration
        In: Proceedings of the conference on Learning and Intelligent OptimizatioN (LION 5)

    Example:
        ``` python
        my_pipe = Hyperpipe('smac_example',
                            optimizer='smac',
                            optimizer_params={"facade": "SMAC4BO",
                                              "wallclock_limit": 60.0*10,  # seconds
                                              "ta_run_limit": 100},  # limit of configurations
                            ...)

        ```

    """
    def __init__(self, facade='SMAC4HPO',
                 run_obj: str = "quality",
                 deterministic: str = "true",
                 wallclock_limit: float = 60.0,
                 intensifier_kwargs: dict = None,
                 rng: int = 42, **kwargs):
        """
        Initialize the object.

        Parameters:
            facade:
                Choice of the SMAC backend strategy, [SMAC4BO, SMAC4HPO, SMAC4AC, BOHB4HPO].

            run_obj:
                Defines the optimization metric.
                When optimizing runtime, cutoff_time is required as well.

            wallclock_limit:
                Maximum amount of wallclock-time used for optimization.

            deterministic:
                If true, SMAC assumes that the target function or algorithm
                is deterministic (the same static seed of 0
                is always passed to the function/algorithm).
                If false, different random seeds are passed
                to the target function/algorithm.

            intensifier_kwargs:
                Dict for intensifier settings.

            rng:
                Random seed of SMAC.facade.

            **kwargs:
                All initial kwargs are passed to SMACs scenario.
                [List of all a vailable parameters](
                https://automl.github.io/SMAC3/master/options.html#scenario).

        """
        super(SMACOptimizer, self).__init__()

        if not __found__:
            msg = "Module smac not found or not installed as expected. " \
                  "Please install the smac/requirements.txt PHOTONAI provides."
            logger.error(msg)
            raise ModuleNotFoundError(msg)

        self.run_obj = run_obj
        self.deterministic = deterministic
        self.wallclock_limit = wallclock_limit
        self.kwargs = kwargs

        if facade in ["SMAC4BO", SMAC4BO, "SMAC4AC", SMAC4AC, "SMAC4HPO", SMAC4HPO, "BOHB4HPO", BOHB4HPO]:
            if type(facade) == str:

                self.facade = eval(facade)
            else:
                self.facade = facade
        else:
            msg = "SMAC.facade {} not known. Please use one of ['SMAC4BO', 'SMAC4AC', 'SMAC4HPO']."
            logger.error(msg.format(str(facade)))
            raise ValueError(msg.format(str(facade)))

        self.rng = rng
        if not intensifier_kwargs:
            self.intensifier_kwargs = {}
        else:
            self.intensifier_kwargs = intensifier_kwargs

        self.cspace = ConfigurationSpace()  # hyperparameter space for SMAC
        self.switch_optiones = {}
        self.hyperparameters = []

        self.maximize_metric = False
        self.constant_dictionary = {}

    def prepare(self, pipeline_elements: list, maximize_metric: bool, objective_function: Callable):
        """
        Initializes the SMAC Optimizer.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

            objective_function:
                The cost or objective function.

        """
        self.cspace = ConfigurationSpace()  # build space
        self._build_smac_space(pipeline_elements)
        if self.constant_dictionary:
            msg = "PHOTONAI has detected some one-valued params in your hyperparameters. Pleas use the kwargs for " \
                  "constant values. This run ignores following settings: " + str(self.constant_dictionary.keys())
            logger.warning(msg)
            warnings.warn(msg)
        self.maximize_metric = maximize_metric

        scenario_dict = self.kwargs
        scenario_dict.update({"run_obj":self.run_obj,
                              "deterministic":self.deterministic,
                              "wallclock_limit":self.wallclock_limit,
                              "cs":self.cspace,
                              "limit_resources": False})

        scenario = Scenario(scenario_dict)

        def smac_objective_function(current_config):
            current_config = {k: current_config[k] for k in current_config if (current_config[k] and 'algos' not in k)}
            return objective_function(current_config)

        self.smac = self.facade(scenario=scenario,
                                intensifier_kwargs=self.intensifier_kwargs,
                                rng=self.rng,
                                tae_runner=smac_objective_function)

    def optimize(self):
        """Start optimization process."""
        self.smac.optimize()

    def _build_smac_space(self, pipeline_elements: list):
        """
        Build the entire SMAC hyperparameter space.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

        """
        for pipe_element in pipeline_elements:
            # build conditions for switch elements
            if pipe_element.__class__.__name__ == 'Switch':
                algorithm_options = {}
                for algo in pipe_element.elements:
                    algo_params = []  # hyper params corresponding to "algo"
                    for name, value in algo.hyperparameters.items():
                        smac_param = self._convert_photonai_to_smac_param(value, (
                                pipe_element.name + "__" + name))  # or element.name__algo.name__name
                        algo_params.append(smac_param)
                    algorithm_options[(pipe_element.name + "__" + algo.name)] = algo_params

                algos = CategoricalHyperparameter(pipe_element.name + "__algos", choices=algorithm_options.keys())

                self.switch_optiones[pipe_element.name + "__algos"] = algorithm_options.keys()

                self.cspace.add_hyperparameter(algos)
                for algo, params in algorithm_options.items():
                    for param in params:
                        cond = InCondition(child=param, parent=algos, values=[algo])
                        self.cspace.add_hyperparameter(param)
                        self.cspace.add_condition(cond)
                continue

            if hasattr(pipe_element, 'hyperparameters'):
                for name, value in pipe_element.hyperparameters.items():
                    self.hyperparameters.append(name)
                    # if we only have one value we do not need to optimize
                    if isinstance(value, list) and len(value) < 2:
                        self.constant_dictionary[name] = value[0]
                        continue
                    if isinstance(value, PhotonCategorical) and len(value.values) < 2:
                        self.constant_dictionary[name] = value.values[0]
                        continue
                    smac_param = self._convert_photonai_to_smac_param(value, name)
                    if smac_param is not None:
                        self.cspace.add_hyperparameter(smac_param)

    @staticmethod
    def _convert_photonai_to_smac_param(hyperparam: PhotonHyperparam, name: str):
        """
        Helper function: Convert PHOTONAI to SMAC hyperparameter.

        Parameters:
            hyperparam:
                One of photonai.optimization.hyperparameters.

            name:
                Name of hyperparameter.

        """
        if isinstance(hyperparam, PhotonCategorical) or isinstance(hyperparam, BooleanSwitch):
            return CategoricalHyperparameter(name, hyperparam.values)
        elif isinstance(hyperparam, list):
            return CategoricalHyperparameter(name, hyperparam)
        elif isinstance(hyperparam, FloatRange):
            if hyperparam.range_type in ['linspace', 'logspace']:
                return UniformFloatHyperparameter(name, hyperparam.start, hyperparam.stop,
                                                  log=(hyperparam.range_type == 'logspace'))
            msg = str(hyperparam.range_type) + "in your FloatRange is not implemented in SMAC."
            logger.error(msg)
            raise NotImplementedError(msg)
        elif isinstance(hyperparam, IntegerRange):
            return UniformIntegerHyperparameter(name, hyperparam.start, hyperparam.stop)

        msg = "Cannot convert hyperparameter " + str(hyperparam) + ". Supported types: Categorical, IntegerRange," \
                                                                   "FloatRange, list."
        logger.error(msg)
        raise ValueError(msg)



import numpy as np
from sklearn.datasets import load_boston
from sklearn.dummy import DummyRegressor
from sklearn.model_selection import ShuffleSplit

from photonai.base import PipelineElement, Hyperpipe
from photonai.base.photon_pipeline import PhotonPipeline
from photonai.optimization import DummyPerformanceConstraint, MinimumPerformanceConstraint, GridSearchOptimizer
from photonai.optimization.optimization_info import Optimization
from photonai.processing.outer_folds import OuterFoldManager
from photonai.processing.photon_folds import FoldInfo
from photonai.processing.results_structure import MDBOuterFold
from photonai.helper.photon_base_test import PhotonBaseTest


class OuterFoldTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(OuterFoldTests, cls).setUpClass()

    def setUp(self):

        super(OuterFoldTests, self).setUp()
        self.fold_nr_inner_cv = 5
        self.inner_cv = ShuffleSplit(n_splits=self.fold_nr_inner_cv, random_state=42)
        self.outer_cv = ShuffleSplit(n_splits=1, test_size=0.2, random_state=42)
        self.cv_info = Hyperpipe.CrossValidation(inner_cv=self.inner_cv,
                                                 outer_cv=self.outer_cv,
                                                 use_test_set=True,
                                                 test_size=0.2,
                                                 calculate_metrics_per_fold=True,
                                                 calculate_metrics_across_folds=False,
                                                 learning_curves=False,
                                                 learning_curves_cut=None)

        self.X, self.y = load_boston(return_X_y=True)
        self.outer_fold_id = "TestFoldOuter1"
        self.cv_info.outer_folds = {self.outer_fold_id: FoldInfo(0, 1, train, test) for train, test in
                                    self.outer_cv.split(self.X, self.y)}

        self.config_num = 2
        self.optimization_info = Optimization(metrics=['mean_absolute_error', 'mean_squared_error'],
                                              best_config_metric='mean_absolute_error',
                                              optimizer_input='grid_search', optimizer_params={},
                                              performance_constraints=None)
        self.elements = [PipelineElement('StandardScaler'),
                         PipelineElement('PCA', {'n_components': [4, 7]}),
                         PipelineElement('DecisionTreeRegressor', random_state=42)]
        self.pipe = PhotonPipeline([(p.name, p) for p in self.elements])

    def prepare_and_fit(self, outer_fold_man=None):
        if outer_fold_man is None:
            outer_fold = MDBOuterFold(fold_nr=1)
            outer_fold_man = OuterFoldManager(self.pipe, self.optimization_info, self.outer_fold_id, self.cv_info,
                                              result_obj=outer_fold)
        outer_fold_man.fit(self.X, self.y)
        return outer_fold_man

    def test_fit(self):
       self.prepare_and_fit()

    def test_current_best_config(self):

        def check_current_best_config_equality(outer_manager, fold_operation):
            # we know that the first of the two configs is better
            # the value in outer_manager.current_best_config should be the mean value of the best config
            # for train
            self.assertEqual(str(outer_manager.result_object.best_config.get_train_metric(self.optimization_info.best_config_metric,
                                                                                          fold_operation)),
                             str(outer_manager.current_best_config.get_train_metric(self.optimization_info.best_config_metric,
                                                                                    fold_operation)))
            # and for test
            self.assertEqual(str(outer_manager.result_object.best_config.get_test_metric(self.optimization_info.best_config_metric,
                                                                                         fold_operation)),
                             str(outer_manager.current_best_config.get_test_metric(self.optimization_info.best_config_metric,
                                                                                   fold_operation)))

        # if we have calculate_metrics_per_fold = True then
        # we take the mean value for evaluating the current best metric
        self.cv_info.calculate_metrics_across_folds = False
        self.cv_info.calculate_metrics_per_fold = True
        outer_manager_per_fold = self.prepare_and_fit()
        check_current_best_config_equality(outer_manager_per_fold, "mean")

        # if we have calculate_metrics_across_fold = True and cm_per_fold = False, we take the raw value to display
        self.cv_info.calculate_metrics_across_folds = True
        self.cv_info.calculate_metrics_per_fold = False
        outer_manager_across_folds = self.prepare_and_fit()
        check_current_best_config_equality(outer_manager_across_folds, "raw")

    def test_prepare(self):
        self.optimization_info.performance_constraints = [DummyPerformanceConstraint(self.optimization_info.best_config_metric),
                                                          MinimumPerformanceConstraint('mean_squared_error', 75)]
        outer_fold_man = OuterFoldManager(self.pipe, self.optimization_info, self.outer_fold_id, self.cv_info,
                                          result_obj=MDBOuterFold(fold_nr=1))

        outer_fold_man._prepare_optimization()
        outer_fold_man._prepare_data(self.X, self.y)
        # test that performance constraints are copies
        self.assertTrue(outer_fold_man.constraint_objects, list)
        self.assertTrue(len(outer_fold_man.constraint_objects) == 2)
        for ico, copied_object in enumerate(outer_fold_man.constraint_objects):
            self.assertIsNot(self.optimization_info.performance_constraints[ico], copied_object)

        # test that optimizer is prepared and can generated our two configs
        self.assertIsNotNone(outer_fold_man.optimizer)
        self.assertTrue(outer_fold_man.optimizer, GridSearchOptimizer)
        self.assertTrue(len(list(outer_fold_man.optimizer.ask)) == 2)

        # assure that we assured there are no cython leftovers in result tree
        self.assertEqual(len(outer_fold_man.result_object.tested_config_list), 0)

        # test that data is split (we only check y because the split method is already tested, we just make sure it is applied)
        nr_train = len(self.cv_info.outer_folds[self.outer_fold_id].train_indices)
        self.assertTrue(len(outer_fold_man._validation_y) == nr_train)
        nr_test = len(self.cv_info.outer_folds[self.outer_fold_id].test_indices)
        self.assertTrue(len(outer_fold_man._test_y) == nr_test)

        # test that infos are in tree
        self.assertEqual(outer_fold_man.result_object.number_samples_validation, nr_train)
        self.assertEqual(outer_fold_man.result_object.number_samples_test, nr_test)

    def test_eval_final_performance(self):
        # check that best_config_score is copied and not computed
        def case_check(outer_fold_man, operation="mean"):
            # we do so by asserting the values that are in the test set position are EXCACTLY the
            # same as in the values from the validation set in the inner folds (hence, copied, not computed)
            for m_name, m_val in outer_fold_man.result_object.best_config.best_config_score.validation.metrics.items():
                compare_value = outer_fold_man.result_object.best_config.get_test_metric(name=m_name,
                                                                                         operation=operation)
                self.assertAlmostEqual(m_val, compare_value)

            # additionally make sure that there are no predictions and no indices for the test set
            self.assertTrue(len(outer_fold_man.result_object.best_config.best_config_score.validation.indices) == 0)
            self.assertTrue(len(outer_fold_man.result_object.best_config.best_config_score.validation.y_pred) == 0)

        # in case we don't evaluate the test set
        self.cv_info.use_test_set = False
        # todo: metric refactoring update
        # we copy the mean value
        self.cv_info.calculate_metrics_across_folds = False
        self.cv_info.calculate_metrics_per_fold = True
        outer_fold_man = self.prepare_and_fit()
        case_check(outer_fold_man)

        # we still copy the mean value
        self.cv_info.calculate_metrics_across_folds = True
        self.cv_info.calculate_metrics_per_fold = True
        outer_fold_man = self.prepare_and_fit()
        case_check(outer_fold_man)

        # we copy the raw metrics computed across the folds
        self.cv_info.calculate_metrics_across_folds = True
        self.cv_info.calculate_metrics_per_fold = False
        outer_fold_man = self.prepare_and_fit()
        case_check(outer_fold_man, operation="raw")

    def test_save_predictions_and_feature_importances(self):
        # todo: the best config should have feature importances and predictions, all others shouldn't have them
        #  reuse test_save_predictions() and test_feature_importances() from inner_fold_tests.py
        np.random.seed(42)
        outer_fold_man1 = OuterFoldManager(self.pipe, self.optimization_info, self.outer_fold_id, self.cv_info,
                                           result_obj=MDBOuterFold(fold_nr=1))
        self.prepare_and_fit(outer_fold_man1)

        best_config = outer_fold_man1.result_object.best_config
        self.assertTrue(len(best_config.best_config_score.validation.y_true) == len(
            self.cv_info.outer_folds[self.outer_fold_id].test_indices))
        self.assertTrue(len(best_config.best_config_score.validation.y_pred) == len(
            self.cv_info.outer_folds[self.outer_fold_id].test_indices))
        self.assertTrue(len(best_config.best_config_score.feature_importances) == 7)

        for config in outer_fold_man1.result_object.tested_config_list:
            if config.config_nr == best_config.config_nr:
                for fold_i, fold in enumerate(config.inner_folds):
                    self.assertTrue(np.sum(len(fold.validation.y_pred) == 41))
                    self.assertTrue(np.sum(len(fold.feature_importances) == 7))
            else:
                self.assertTrue(np.sum(len(fold.validation.y_pred) for fold in config.inner_folds) == 0)
                self.assertTrue(np.sum(len(fold.validation.probabilities) for fold in config.inner_folds) == 0)
                self.assertTrue(np.sum(len(fold.validation.indices) for fold in config.inner_folds) == 0)
                self.assertTrue(np.sum(len(fold.feature_importances) for fold in config.inner_folds) == 0)

    def test_find_best_config_always_again(self):
        outer_fold_man1 = self.prepare_and_fit()
        outer_fold_man2 = self.prepare_and_fit()

        # we have different entities
        self.assertTrue(outer_fold_man1 is not outer_fold_man2)

        # and they both found the same configuration
        self.assertDictEqual(outer_fold_man1.result_object.best_config.config_dict,
                             outer_fold_man2.result_object.best_config.config_dict)

        # and they both calculated exactly the same values for inner_cv and test set
        self.assertListEqual([str(m) for m in outer_fold_man1.result_object.best_config.metrics_train],
                             [str(m) for m in outer_fold_man2.result_object.best_config.metrics_train])

        self.assertListEqual([str(m) for m in outer_fold_man1.result_object.best_config.metrics_test],
                             [str(m) for m in outer_fold_man2.result_object.best_config.metrics_test])

        self.assertDictEqual(outer_fold_man1.result_object.best_config.best_config_score.validation.metrics,
                             outer_fold_man2.result_object.best_config.best_config_score.validation.metrics,)

        self.assertDictEqual(outer_fold_man1.result_object.best_config.best_config_score.training.metrics,
                             outer_fold_man2.result_object.best_config.best_config_score.training.metrics, )

    def test_fit_dummy(self):
        self.optimization_info.performance_constraints = DummyPerformanceConstraint(self.optimization_info.best_config_metric)
        outer_fold_man = OuterFoldManager(self.pipe, self.optimization_info, self.outer_fold_id, self.cv_info,
                                          result_obj=MDBOuterFold(fold_nr=1))

        outer_fold_man._prepare_optimization()

        # check skipping if no dummy_estimator is given
        outer_fold_man._prepare_data(self.X, self.y)
        outer_fold_man._fit_dummy()
        self.assertIsNone(outer_fold_man.result_object.dummy_results)

        # check for too much dimensions
        outer_fold_man._prepare_data(np.ones((self.X.shape[0], self.X.shape[1], 1)), self.y)
        outer_fold_man._fit_dummy()
        self.assertIsNone(outer_fold_man.result_object.dummy_results)

        # check that dummy result exists with the correct values
        outer_fold_man.dummy_estimator = DummyRegressor()
        outer_fold_man._prepare_data(self.X, self.y)
        outer_fold_man._fit_dummy()

        # for boston housing we expect
        train_values = {'mean_absolute_error': 6.809283403587883, 'mean_squared_error': 86.87340383295755}
        test_values = {'mean_absolute_error': 6.255843525529023, 'mean_squared_error': 75.04543037399255}

        self.assertDictEqual(outer_fold_man.result_object.dummy_results.validation.metrics, test_values)
        self.assertDictEqual(outer_fold_man.result_object.dummy_results.training.metrics, train_values)

        # check that performance constraints are updated
        self.assertTrue(outer_fold_man.constraint_objects[0].threshold ==
                        outer_fold_man.result_object.dummy_results.validation.metrics[self.optimization_info.best_config_metric])



import os
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, PhotonRegistry
from photonai.optimization import IntegerRange

# REGISTER ELEMENT
base_folder = os.path.dirname(os.path.abspath(__file__))
custom_elements_folder = os.path.join(base_folder, '../advanced/custom_elements')

registry = PhotonRegistry(custom_elements_folder=custom_elements_folder)
registry.register(photon_name='MyCustomEstimator',
                  class_str='custom_estimator.CustomEstimator',
                  element_type='Estimator')

registry.register(photon_name='MyCustomTransformer',
                  class_str='custom_transformer.CustomTransformer',
                  element_type='Transformer')

registry.activate()

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('custom_estimator_pipe',
                    optimizer='random_grid_search',
                    optimizer_params={'n_configurations': 2},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=1,
                    project_folder='./tmp/')


# SHOW WHAT IS POSSIBLE IN THE CONSOLE
registry.list_available_elements()

# NOW FIND OUT MORE ABOUT A SPECIFIC ELEMENT
registry.info('MyCustomEstimator')
registry.info('MyCustomTransformer')

my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}, test_disabled=True)

my_pipe += PipelineElement('MyCustomEstimator')


# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



"""
===========================================================
Project: PHOTON Model Wrapper
===========================================================
Description
-----------
Wrapper for sample pairing

Version
-------
Created:        25-03-2019
Last updated:   25-03-2019


Author
------
Tim Hahn & Nils Winter
Translationale Psychiatrie
Universitaetsklinikum Muenster
"""

import itertools
import math
import random

import numpy as np
from scipy.spatial.distance import pdist
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler

from photonai.helper.helper import PhotonDataHelper
from photonai.photonlogger.logger import logger


class SamplePairingBase(BaseEstimator, TransformerMixin):

    @staticmethod
    def _stirling(n):
        # http://en.wikipedia.org/wiki/Stirling%27s_approximation
        return math.sqrt(2 * math.pi * n) * (n / math.e) ** n

    @staticmethod
    def _calculate_number_of_possible_combinations(n, r):
        """
        Get number of combinations of r from n items (e.g. n=10, r=2--> pairs from 0 to 10
        :param n: n items to draw from
        :param r: number of items to draw each time
        :return: approximate number of combinations
        """
        try:
            stirling_n = SamplePairingBase._stirling(n)
            stirling_r = SamplePairingBase._stirling(r)
            stirling_n_r = SamplePairingBase._stirling(n - r)

            if n > 20:
                return stirling_n / stirling_r / stirling_n_r
            else:
                return math.factorial(n) / math.factorial(r) / math.factorial(n - r)
        except:
            return -1

    @staticmethod
    def random_pair_generator(sample_indices, rand_seed, draw_limit):
        """Return a list of random pairs from a list of items."""
        random.seed(rand_seed)

        pairs = list()

        for i in range(draw_limit):
            n_attempts = 0
            while True:
                pair = tuple(sorted(random.sample(sample_indices, 2)))
                if pair not in pairs:
                    pairs.append(pair)
                    break
                elif n_attempts > draw_limit:
                    # prevent while loop from continuing infinitely
                    # stop in case the number of attempts exceeds the draw limit
                    # this is just an arbitrary threshold
                    break
                n_attempts += 1
        return pairs

    @staticmethod
    def nearest_pair_generator(X, distance_metric, draw_limit):
        n_samples = X.shape[0]
        scaler = StandardScaler()
        X = scaler.fit_transform(X)
        distance_indices = np.argsort(pdist(X, distance_metric))
        triu_indices = np.triu_indices(n_samples, k=1)

        pairs = list()
        for i in range(draw_limit):
            if i > len(distance_indices) - 1:
                break
            else:
                pairs.append((triu_indices[0][distance_indices[i]], (triu_indices[1][distance_indices[i]])))
        return pairs

    def _get_pairs(self, X, draw_limit, rand_seed, distance_metric, generator='random_pair'):
        """
        :param X: data array
        :param draw_limit: in case the full number of combinations is > 10k, how many to draw?
        :param rand_seed: sets seed for random sampling of combinations (for reproducibility)
        :param generator: method with which to obtain sample pairs (samples until draw_limit is reached or generator stops)
                            'random_pair': sample randomly from all pairs
                            'nearest_pair': get most similar pairs
        :param distance metric: if generator is 'nearest_pair', this will set the distance metric to obtained similarity
        :return: list of tuples indicating which samples to merge
        """

        sample_indices = range(X.shape[0])

        # limit the number of new samples generated if all combinations > draw_limit
        n_combinations = self._calculate_number_of_possible_combinations(n=len(sample_indices), r=2)

        if n_combinations > draw_limit or n_combinations == -1:
            if generator == 'random_pair':
                return self.random_pair_generator(sample_indices=sample_indices,
                                                  rand_seed=rand_seed,
                                                  draw_limit=draw_limit)
            elif generator == 'nearest_pair':
                return self.nearest_pair_generator(X=X,
                                                   distance_metric=distance_metric,
                                                   draw_limit=draw_limit)
            else:
                raise NotImplementedError("{} is not supported. Possible options: 'random_pair', 'nearest_pair")
        else:
            # get all combinations of samples
            return list(itertools.combinations(sample_indices, 2))

    def _return_samples(self, X, y, kwargs, generator, distance_metric, draw_limit, rand_seed):
        """
                Generates "new samples" by computing the mean between all or n_draws pairs of existing samples and appends them to X
                The target for each new sample is computed as the mean between the constituent targets
                :param X: data
                :param y: targets (optional)
                :param draw_limit: in case the full number of combinations is > 10k, how many to draw?
                :param rand_seed: sets seed for random sampling of combinations (for reproducibility only)
                :return: X_new: X and X_augmented; (y_new: the correspoding targets)
                """
        # generate combinations
        pairs = self._get_pairs(X=X,
                                generator=generator,
                                distance_metric=distance_metric,
                                draw_limit=draw_limit,
                                rand_seed=rand_seed)

        # compute mean over sample pairs
        X_new = np.empty([len(pairs), X.shape[1]])
        i = 0
        for pair in pairs:
            X_new[i] = np.mean([X[pair[0]], X[pair[1]]], axis=0)
            i += 1

        # add augmented samples to existing data
        X_new = np.concatenate((X, X_new), axis=0)

        # get the corresponding targets and kwargs
        if kwargs:
            for name, kwarg in kwargs.items():
                kwarg_new = np.empty(len(pairs))
                for i, pair in enumerate(pairs):
                    kwarg_new[i] = np.mean([kwargs[name][pair[0]], kwargs[name][pair[1]]])
                kwargs[name] = np.concatenate((np.asarray(kwargs[name]), kwarg_new))

        if y is not None:
            y_aug = np.empty(len(pairs))
            i = 0
            for pair in pairs:
                y_aug[i] = np.mean([y[pair[0]], y[pair[1]]])
                i += 1
            # add augmented samples to existing data
            y_new = np.concatenate((y, y_aug))
            return X_new, y_new, kwargs
        else:
            return X_new, None, kwargs


class SamplePairingRegression(SamplePairingBase):
    _estimator_type = "transformer"

    def __init__(self, generator='random_pair', distance_metric='euclidean', draw_limit=10000, random_state=45):

        self.needs_covariates = True
        self.needs_y = True
        self.generator = generator
        self.distance_metric = distance_metric
        self.draw_limit = draw_limit
        self.random_state = random_state

    def fit(self, X, y=None, **kwargs):
        return self

    def transform(self, X, y=None, **kwargs):
        """
        Generates "new samples" by computing the mean between all or n_draws pairs of existing samples and appends them to X
        The target for each new sample is computed as the mean between the constituent targets
        :param X: data
        :param y: targets (optional)
        :param draw_limit: in case the full number of combinations is > 10k, how many to draw?
        :param rand_seed: sets seed for random sampling of combinations (for reproducibility only)
        :return: X_new: X and X_augmented; (y_new: the correspoding targets)
        """
        return self._return_samples(X, y, kwargs, self.generator, self.distance_metric, self.draw_limit, self.random_state)


class SamplePairingClassification(SamplePairingBase):
    _estimator_type = "transformer"

    def __init__(self, generator='random_pair', distance_metric='euclidean', draw_limit=10000, random_state=45,
                 balance_classes=True):

        self.needs_covariates = True
        self.needs_y = True
        self.generator = generator
        self.distance_metric = distance_metric
        self.draw_limit = draw_limit
        self.random_state = random_state
        self.balance_classes = balance_classes

    def fit(self, X, y=None, **kwargs):
        return self

    def transform(self, X, y=None, **kwargs):
        """
        Generates "new samples" by computing the mean between all or n_draws pairs of existing samples and appends them to X
        The target for each new sample is computed as the mean between the constituent targets
        :param X: data
        :param y: targets (optional)
        :param draw_limit: in case the full number of combinations is > 10k, how many to draw?
        :param rand_seed: sets seed for random sampling of combinations (for reproducibility only)
        :return: X_new: X and X_augmented; (y_new: the correspoding targets)
        """

        logger.debug("Pairing " + str(self.draw_limit) + " samples...")

        # ensure class balance in the training set if balance_classes is True
        unique_classes = np.unique(y)
        n_pairs = list()
        for label in unique_classes:
            if self.balance_classes:
                n_pairs.append(self.draw_limit - np.sum(y == label))
            else:
                n_pairs.append(self.draw_limit)

        # run get_samples for each class independently
        X_extended = list()
        y_extended = list()
        kwargs_extended = dict()

        for label, limit in zip(unique_classes, n_pairs):
            X_new_class, y_new_class, kwargs_new_class = self._return_samples(X[y == label], y[y == label],
                                                                              PhotonDataHelper.index_dict(kwargs,
                                                                                                          y == label),
                                                                              generator=self.generator,
                                                                              distance_metric=self.distance_metric,
                                                                              draw_limit=limit,
                                                                              rand_seed=self.random_state)

            X_extended.extend(X_new_class)
            y_extended.extend(y_new_class)

            # get the corresponding kwargs
            if kwargs:
                kwargs_extended = PhotonDataHelper.join_dictionaries(kwargs_extended, kwargs_new_class)

        return X_extended, y_extended, kwargs_extended



import unittest
import numpy as np
import warnings

from photonai.optimization import FloatRange, IntegerRange, Categorical, BooleanSwitch
from photonai.optimization.hyperparameters import NumberRange


class HyperparameterBaseTest(unittest.TestCase):

    def setUp(self):
        """
        Set default start setting for all tests.
        """
        self.intger_range = IntegerRange(2,6)
        self.float_range = FloatRange(0.1, 5.7)
        self.cateogrical_truth = ["a","b","c","d","e","f","g","h"]
        self.categorical = Categorical(self.cateogrical_truth)
        self.bool = BooleanSwitch()

    def test_rand_success(self):

        for _ in range(100):
            self.assertIn(self.intger_range.get_random_value(), list(range(2,6)))

            self.assertGreaterEqual(self.float_range.get_random_value(), 0.1)
            self.assertLess(self.float_range.get_random_value(), 5.7)

            self.assertIn(self.categorical.get_random_value(), self.cateogrical_truth)

            self.assertIn(self.bool.get_random_value(), [True, False])

        self.float_range.transform()
        self.intger_range.transform()

        for _ in range(100):
            self.assertIn(self.intger_range.get_random_value(definite_list=True), self.intger_range.values)
            self.assertIn(self.float_range.get_random_value(definite_list=True), self.float_range.values)

    def test_domain(self):

        self.float_range.transform()
        self.intger_range.transform()
        self.assertListEqual(self.intger_range.values, list(np.arange(2,6)))
        self.assertListEqual(self.float_range.values, list(np.linspace(0.1, 5.7, dtype=np.float64)))

        big_float_range = FloatRange(-300.57, np.pi*4000)
        big_float_range.transform()
        self.assertListEqual(big_float_range.values, list(np.linspace(-300.57, np.pi*4000)))
        self.assertListEqual(self.categorical.values, ["a","b","c","d","e","f","g","h"])
        self.assertListEqual(self.bool.values, [True, False])

    def test_rand_error(self):
        with self.assertRaises(ValueError):
            self.intger_range.get_random_value(definite_list=True)
        with self.assertRaises(ValueError):
            self.float_range.get_random_value(definite_list=True)
        with self.assertRaises(NotImplementedError):
            self.categorical.get_random_value(definite_list=False)
        with self.assertRaises(NotImplementedError):
            self.categorical.get_random_value(definite_list=False)

    def test_categorical(self):
        self.assertEqual(self.categorical[2], self.cateogrical_truth[2])


class NumberRangeTest(unittest.TestCase):

    def setUp(self):
        """
        Set default start setting for all tests.
        """
        self.start = np.random.randint(0, 5)
        self.end = np.random.randint(100, 105)

    def test_NumberRange(self):
        """
        Test for class IntegerRange and FloatRange.
        """
        dtypes = {int: np.int32, float: np.float32}

        expected_linspace = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        number_range_linspace = IntegerRange(start=1, stop=9, num=9, range_type="linspace")
        number_range_linspace.transform()
        self.assertListEqual(expected_linspace, number_range_linspace.values)

        expected_geomspace = [1, 10, 100, 1000, 10000]
        number_range_geomspace = IntegerRange(1, 10000, num=5, range_type="geomspace")
        number_range_geomspace.transform()
        self.assertListEqual(expected_geomspace, number_range_geomspace.values)

        number_range_range = IntegerRange(self.start, self.end, step=2, range_type="range")
        number_range_range.transform()
        self.assertListEqual(number_range_range.values, list(np.arange(self.start, self.end, 2)))

        number_range_logspace = FloatRange(-1, 1, num=50, range_type='logspace')
        number_range_logspace.transform()
        np.testing.assert_array_almost_equal(number_range_logspace.values,  np.logspace(-1, 1, num=50))

        # error tests
        with self.assertRaises(ValueError):
            number_range = IntegerRange(start=0, stop=self.end, range_type="geomspace")
            number_range.transform()

        number_range = FloatRange(start=0.1, stop=self.end, range_type="geomspace")
        number_range.transform()

        with self.assertRaises(ValueError):
            number_range = IntegerRange(start=1, stop=15, range_type="logspace")
            number_range.transform()
        number_range = FloatRange(start=0.1, stop=self.end, range_type="logspace")
        number_range.transform()

        with self.assertRaises(ValueError):
            IntegerRange(start=self.start, stop=self.end, range_type="ownspace")

    def test_false_range_type(self):
        with self.assertRaises(ValueError):
            float_range = FloatRange(1.0, 5.2, range_type='normal_distributed')
            float_range.transform()

class HyperparameterOtherTest(unittest.TestCase):

    def test_categorical(self):
        """
        Test for class Categorical.
        """
        items = "Lorem ipsum dolor sit amet consetetur sadipscing elitr".split(" ")
        categorical = Categorical(values=items)
        self.assertEqual(categorical[2], "dolor")
        self.assertListEqual(categorical[2:5], items[2:5])
        self.assertListEqual(categorical.values, items)

    def test_boolean_switch(self):
        """
        Test for class BooleanSwitch.
        """
        boolean_switch = BooleanSwitch()
        self.assertListEqual(boolean_switch.values, [True, False])

    def test_start_stop_problem(self):
        with warnings.catch_warnings(record=True) as w:
            integer_range = IntegerRange(2, 1)
            integer_range.transform()
            assert any("NumberRange or one of its subclasses is empty" in s for s in [e.message.args[0] for e in w])

    def test_dtypes(self):
        complex_range = NumberRange(1, 5, num_type=complex, range_type="range")
        complex_range.transform()
        str_range = NumberRange(1, 2, num_type=np.bool_, range_type="range")
        str_range.transform()



import types
import unittest
from functools import reduce
import operator
from inspect import signature

from photonai.base import PipelineElement, Switch, Branch, Hyperpipe
from photonai.optimization import GridSearchOptimizer, RandomGridSearchOptimizer, IntegerRange
from photonai.optimization.base_optimizer import PhotonSlaveOptimizer, PhotonMasterOptimizer

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold, ShuffleSplit


class GridSearchOptimizerTest(unittest.TestCase):

    def setUp(self):
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}),
                                  PipelineElement("SVC")]
        self.optimizer = GridSearchOptimizer()
        self.optimizer_name = 'grid_search'
        self.optimizer_params = None

    def create_hyperpipe(self):
        self.hyperpipe = Hyperpipe('optimizer_test',
                                   project_folder='./tmp',
                                   metrics=['accuracy'],
                                   best_config_metric='accuracy',
                                   inner_cv=KFold(n_splits=2),
                                   outer_cv=ShuffleSplit(n_splits=2),
                                   optimizer=self.optimizer_name,
                                   optimizer_params=self.optimizer_params,
                                   verbosity=0)

    def test_run(self):
        self.create_hyperpipe()
        for p in self.pipeline_elements:
            self.hyperpipe += p
        X, y = load_breast_cancer(return_X_y=True)
        self.hyperpipe.fit(X, y)

    def test_all_functions_available(self):
        """Test existence of functions and parameters ->  .ask() .tell() .prepare()."""
        self.assertTrue(hasattr(self.optimizer, 'prepare'))
        self.assertListEqual(list(signature(self.optimizer.prepare).parameters.keys()),
                             ['pipeline_elements', 'maximize_metric'])
        self.assertTrue(hasattr(self.optimizer, 'tell'))
        self.assertListEqual(list(signature(self.optimizer.tell).parameters.keys()), ['config', 'performance'])
        self.assertTrue(hasattr(self.optimizer, 'ask'))

    def test_all_attributes_available(self):
        """Test for .ask and .param_grid attribute. .ask is important for next configuration that should be tested."""
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        self.assertIsInstance(self.optimizer.ask, types.GeneratorType)

    def test_ask(self):
        """Test general functionality of .ask()."""
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        ask_list = list(self.optimizer.ask)
        self.assertIsInstance(ask_list, list)
        self.assertSetEqual(set([str(type(a)) for a in ask_list]), {"<class 'dict'>"})
        generated_elements = reduce(operator.concat, [list(a.keys()) for a in ask_list])
        self.assertIn("PCA__n_components", generated_elements)
        return generated_elements

    def test_ask_advanced(self):
        """Test advanced functionality of .ask()."""
        branch = Branch('branch')
        branch += PipelineElement('PCA')
        branch += PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['rbf', 'sigmoid']})
        pipe_switch = Switch('switch', [PipelineElement("StandardScaler"), PipelineElement("MaxAbsScaler")])
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)},
                                                  test_disabled=True),
                                  pipe_switch,
                                  branch,
                                  Switch('Switch_in_switch', [branch, pipe_switch])]
        generated_elements = self.test_ask()
        self.assertIn("PCA__n_components", generated_elements)
        self.assertIn("Switch_in_switch__current_element", generated_elements)
        self.assertIn("branch__SVC__C", generated_elements)
        self.assertIn("branch__SVC__kernel", generated_elements)
        self.assertIn("switch__current_element", generated_elements)


class RandomGridSearchOptimizerTest(GridSearchOptimizerTest):

    def setUp(self):
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}),
                                  PipelineElement("SVC")]
        self.optimizer = RandomGridSearchOptimizer()
        self.optimizer_name = 'random_grid_search'
        self.optimizer_params = None

    def test_parameter_k(self):
        """Test for parameter n_configuration and k."""
        self.optimizer = RandomGridSearchOptimizer(n_configurations=3)
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        self.assertEqual(len(self.optimizer.param_grid), 3)
        self.optimizer = RandomGridSearchOptimizer(n_configurations=500)
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        self.assertEqual(len(self.optimizer.param_grid), 15)


class BaseOptimizerTests(unittest.TestCase):

    @staticmethod
    def test_slave_interface():
        opt = PhotonSlaveOptimizer()
        opt.prepare(list(), True)
        opt.ask()
        opt.tell(dict(), float())

    @staticmethod
    def test_master_interface():
        opt = PhotonMasterOptimizer()
        opt.prepare(list(), True, None)
        opt.optimize()



import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold
from sklearn.base import BaseEstimator, ClassifierMixin
from photonai.base import Hyperpipe, PipelineElement


class AdditionalDataWrapper(BaseEstimator, ClassifierMixin):

    def __init__(self):
        self.needs_covariates = True

    def fit(self, X, y, **kwargs):
        if "true_predictions" in kwargs:
            print("Found additional data")
        return self

    def predict(self, X, **kwargs):
        y_true = kwargs["true_predictions"]
        assert X.shape[0] == len(y_true)
        return y_true

    def save(self):
        return None


my_pipe = Hyperpipe('additional_data_pipe',
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=1,
                    project_folder='./tmp/')

my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement.create("CustomWrapper", AdditionalDataWrapper(), hyperparameters={})

X, y = load_breast_cancer(return_X_y=True)
my_pipe.fit(X, y, true_predictions=np.array(y))



import os
import shutil
import uuid
import warnings
from prettytable import PrettyTable

import numpy as np
import pandas as pd
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold
from sklearn.base import BaseEstimator, TransformerMixin

from photonai.base import Hyperpipe, OutputSettings
from photonai.base import PipelineElement, Stack, Switch
from photonai.optimization import IntegerRange, FloatRange
from photonai.processing import ResultsHandler
from photonai.helper.photon_base_test import PhotonBaseTest
from photonai.processing.results_structure import MDBHyperpipe


class ResultsHandlerTest(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(ResultsHandlerTest, cls).setUpClass()

        cls.files = ['best_config_predictions.csv',
                     'photon_result_file.json',
                     'photon_summary.txt',
                     'photon_best_model.photon']

        cls.results_folder = cls.tmp_folder_path
        cls.output_settings = OutputSettings(save_output=True)

        cls.mongodb_path = 'mongodb://localhost:27017/photon_results'

        cls.ss_pipe_element = PipelineElement('StandardScaler')
        cls.pca_pipe_element = PipelineElement('PCA', {'n_components': [1, 2]}, random_state=42)
        cls.svc_pipe_element = PipelineElement('SVC', {'C': [0.1], 'kernel': ['linear']},  # 'rbf', 'sigmoid']
                                               random_state=42)

        cls.inner_cv_object = KFold(n_splits=3)
        cls.metrics = ["accuracy", 'recall', 'precision']
        cls.best_config_metric = "accuracy"
        cls.pipe_name = str(uuid.uuid4())
        cls.hyperpipe = Hyperpipe(cls.pipe_name, inner_cv=cls.inner_cv_object,
                                  metrics=cls.metrics,
                                  best_config_metric=cls.best_config_metric,
                                  outer_cv=KFold(n_splits=2),
                                  project_folder=cls.results_folder,
                                  output_settings=cls.output_settings,
                                  verbosity=0)
        cls.hyperpipe += cls.ss_pipe_element
        cls.hyperpipe += cls.pca_pipe_element
        cls.hyperpipe.add(cls.svc_pipe_element)

        dataset = load_breast_cancer()
        cls.__X = dataset.data
        cls.__y = dataset.target

        cls.hyperpipe.fit(cls.__X, cls.__y)
        
    def test_write_convenience_files(self):
        """
        Output creation testing. Only write if output_settings.save_output == True
        """
        for file in self.files:
            self.assertTrue(os.path.isfile(os.path.join(self.hyperpipe.output_settings.results_folder, file)))

        # correct rows
        with open(os.path.join(self.hyperpipe.output_settings.results_folder, 'best_config_predictions.csv')) as f:
            self.assertEqual(sum([outer_fold.number_samples_test
                                  for outer_fold in self.hyperpipe.results.outer_folds]),
                             sum(1 for _ in f)-1)

        shutil.rmtree(self.tmp_folder_path, ignore_errors=True)
        self.hyperpipe.output_settings = OutputSettings(save_output=False)
        self.hyperpipe.fit(self.__X, self.__y)
        self.assertFalse(os.path.exists(self.hyperpipe.output_settings.results_folder))

    def test_summary(self):
        """
        Check content of photon_summary.txt. Adjustment with hyperpipe.result.
        """

        # todo: check appropriately
        pass

    def test_save_backmapping_weird_format(self):
        self.hyperpipe.fit(self.__X, self.__y)
        weird_format_fake = ('string', ['array'])
        # should be saved as pickle
        self.hyperpipe.results_handler.save_backmapping('weird_format', weird_format_fake)
        expected_file = os.path.join(self.hyperpipe.output_settings.results_folder, 'weird_format.p')
        self.assertTrue(os.path.isfile(expected_file))

    def test_save_backmapping_csv(self):
        """
        Check dimension of feature backmapping equals input dimensions for less than 1000 features.
        """
        backmapping = np.loadtxt(os.path.join(self.hyperpipe.output_settings.results_folder,
                                 'optimum_pipe_feature_importances_backmapped.csv'), delimiter=',')
        self.assertEqual(np.shape(self.__X)[1], backmapping.size)

    def test_save_backmapping_npz(self):
        """
        Check dimension of feature backmapping equals input dimensions for more than 1000 features.
        """
        # run another hyperpipe with more than 1000 features
        # use np.tile to copy features until at least 1000 features are reached
        X = np.tile(self.__X, (1, 35))
        self.hyperpipe.fit(X, self.__y)
        npzfile = np.load(os.path.join(self.hyperpipe.output_settings.results_folder,
                                       'optimum_pipe_feature_importances_backmapped.npz'))
        self.assertEqual(len(npzfile.files), 1)
        backmapping = npzfile[npzfile.files[0]]
        self.assertEqual(np.shape(X)[1], backmapping.size)

    def test_save_backmapping_stack(self):
        # build hyperpipe with stack first
        self.hyperpipe = Hyperpipe(self.pipe_name, inner_cv=self.inner_cv_object,
                                   metrics=self.metrics,
                                   best_config_metric=self.best_config_metric,
                                   outer_cv=KFold(n_splits=2),
                                   output_settings=self.output_settings,
                                   project_folder=self.results_folder,
                                   verbosity=1)
        self.hyperpipe += self.ss_pipe_element
        self.stack = Stack("myStack")
        self.stack += PipelineElement("MinMaxScaler")
        self.stack += self.pca_pipe_element
        self.hyperpipe += self.stack
        self.hyperpipe.add(self.svc_pipe_element)

        self.output_settings.save_output = True
        self.hyperpipe.fit(self.__X, self.__y)

        f_name = os.path.join(self.hyperpipe.output_settings.results_folder,
                              'optimum_pipe_feature_importances_backmapped.csv')
        self.assertTrue(not os.path.isfile(f_name))

    def test_save_backmapping_reduced_dimension_without_inverse(self):
        class RD(BaseEstimator, TransformerMixin):

            def fit(self, X, y, **kwargs):
                pass

            def fit_transform(self, X, y=None, **fit_params):
                return self.transform(X)

            def transform(self, X):
                return X[:, :3]

        trans = PipelineElement.create('MyTransformer', base_element=RD(), hyperparameters={})

        self.hyperpipe = Hyperpipe(self.pipe_name, inner_cv=self.inner_cv_object,
                                   metrics=self.metrics,
                                   best_config_metric=self.best_config_metric,
                                   outer_cv=KFold(n_splits=2),
                                   project_folder=self.results_folder,
                                   output_settings=self.output_settings,
                                   verbosity=1)
        self.hyperpipe += trans
        self.hyperpipe.add(self.svc_pipe_element)

        self.output_settings.save_output = True

        f_name = os.path.join(self.results_folder, 'optimum_pipe_feature_importances_backmapped.csv')
        self.assertTrue(not os.path.isfile(f_name))

    def test_get_feature_importances(self):
        self.hyperpipe.elements[-1] = PipelineElement('LinearSVC')
        self.hyperpipe.fit(self.__X, self.__y)
        f_imps = self.hyperpipe.results_handler.get_importance_scores()
        self.assertIsNotNone(f_imps)

    def test_pass_through_plots(self):
        """
        Test for plot functions. Only passing test, no quality testing.
        """
        self.assertIsNone(self.hyperpipe.results_handler.plot_optimizer_history(metric='accuracy', type='scatter'))
        self.assertIsNone(self.hyperpipe.results_handler.eval_mean_time_components())

    def test_empty_results(self):
        res_handler = ResultsHandler()
        with self.assertRaises(ValueError):
            res_handler.get_test_predictions('test_predictions')
        with self.assertRaises(ValueError):
            res_handler.get_validation_predictions('test_predictions')

    def test_save_all_learning_curves(self):
        """
        Test count of saved learning curve files.
        """
        hyperpipe = Hyperpipe(self.pipe_name, inner_cv=self.inner_cv_object,
                              learning_curves=True,
                              metrics=self.metrics,
                              best_config_metric=self.best_config_metric,
                              outer_cv=KFold(n_splits=2),
                              project_folder=self.results_folder,
                              output_settings=self.output_settings,
                              verbosity=2)
        hyperpipe += self.ss_pipe_element
        hyperpipe += self.pca_pipe_element
        hyperpipe.add(self.svc_pipe_element)
        hyperpipe.fit(self.__X, self.__y)
        results_handler = hyperpipe.results_handler
        results_handler.save_all_learning_curves()
        config_num = len(hyperpipe.results.outer_folds[0].tested_config_list)
        target_file_num = 2 * config_num * hyperpipe.cross_validation.outer_cv.n_splits
        curves_folder = hyperpipe.output_settings.results_folder + '/learning_curves/'
        true_file_num = len([name for name in os.listdir(curves_folder)
                             if os.path.isfile(os.path.join(curves_folder, name))])
        self.assertEqual(target_file_num, true_file_num)

        # test outer fold plots
        results_handler.plot_learning_curves_outer_fold(0)

    def test_load_from_file_and_mongodb(self):
        hyperpipe = self.hyperpipe.copy_me()
        hyperpipe.output_settings.mongodb_connect_url = self.mongodb_path
        hyperpipe.fit(self.__X, self.__y)

        results_file = os.path.join(hyperpipe.output_settings.results_folder, "photon_result_file.json")
        my_result_handler = ResultsHandler()
        my_result_handler.load_from_file(results_file)
        self.assertIsInstance(my_result_handler.results, MDBHyperpipe)

        my_mongo_result_handler = ResultsHandler()
        my_mongo_result_handler.load_from_mongodb(pipe_name=hyperpipe.name,
                                                  mongodb_connect_url=self.mongodb_path)
        self.assertIsInstance(my_mongo_result_handler.results, MDBHyperpipe)
        self.assertTrue(my_mongo_result_handler.results.name == hyperpipe.name)

        # write again to mongodb
        hyperpipe.fit(self.__X, self.__y)
        with warnings.catch_warnings(record=True) as w:
            my_mongo_result_handler.load_from_mongodb(pipe_name=hyperpipe.name,
                                                      mongodb_connect_url=self.mongodb_path)
            assert any("Found multiple hyperpipes with that name." in s for s in [e.message.args[0] for e in w])

        with self.assertRaises(FileNotFoundError):
            my_result_handler.load_from_mongodb(pipe_name='any_weird_name_1238934384834234892384382',
                                                mongodb_connect_url=self.mongodb_path)

    def test_get_performance_table(self):

        performance_table = self.hyperpipe.results_handler.get_performance_table()

        self.assertListEqual(list(performance_table.columns), ["best_config", "fold", "n_train", "n_validation"] +
                             self.metrics +
                             [x+"_sem" for x in self.metrics])

        for i in range(2):
            self.assertEqual(int(performance_table.iloc[i]["fold"]), i+1)
            for key, value in self.hyperpipe.results_handler.results.outer_folds[i].\
                    best_config.best_config_score.validation.metrics.items():
                self.assertEqual(performance_table.iloc[i][key], value)
            self.assertEqual(self.hyperpipe.results_handler.results.outer_folds[i].best_config.
                             best_config_score.number_samples_training,
                             performance_table.iloc[i]["n_train"])
            self.assertEqual(self.hyperpipe.results_handler.results.outer_folds[i].best_config.
                             best_config_score.number_samples_validation,
                             performance_table.iloc[i]["n_validation"])
        self.assertEqual(performance_table.iloc[2].isnull().all(), False)

        self.assertIsInstance(performance_table, pd.DataFrame)

    def test_get_methods(self):
        self.hyperpipe.results_handler.get_methods()

    def test_float_labels_with_mongo(self):
        """
        This test was added for a bug with float labels and saving to mongoDB.
        """
        local_y = self.__y.astype(float)
        self.hyperpipe.output_settings.mongodb_connect_url = self.mongodb_path
        self.hyperpipe.fit(self.__X, local_y)

    def test_get_mean_of_best_validation_configs_per_estimator(self):
        hyperpipe = Hyperpipe('compare_estimators',
                              inner_cv=KFold(n_splits=2, shuffle=True),
                              outer_cv=KFold(n_splits=2, shuffle=True),
                              metrics=['mean_squared_error'],
                              best_config_metric='mean_squared_error',
                              optimizer='switch',
                              optimizer_params={'name': 'random_search', 'n_configurations': 3},
                              project_folder='./tmp',
                              verbosity=0)
        hyperpipe += self.ss_pipe_element

        # compare different learning algorithms in an OR_Element
        estimators = Switch('estimator_selection')
        estimators += PipelineElement('RandomForestRegressor',
                                      hyperparameters={'min_samples_split': IntegerRange(2, 4)})
        estimators += PipelineElement('SVR', hyperparameters={'C': FloatRange(0.5, 25)})

        hyperpipe += estimators
        hyperpipe.fit(self.__X[:150], self.__y[:150])

        output = hyperpipe.results_handler.get_mean_of_best_validation_configs_per_estimator()
        output2 = hyperpipe.results_handler.get_n_best_validation_configs_per_estimator()

        self.assertTrue(isinstance(output, PrettyTable))
        self.assertTrue(isinstance(output2, dict))




from photonai.helper.photon_base_test import PhotonBaseTest
from photonai.base import PipelineElement, Hyperpipe, ParallelBranch
from photonai.base.cache_manager import CacheManager
from photonai.base.photon_pipeline import PhotonPipeline

from sklearn.model_selection import KFold
from sklearn.base import BaseEstimator
import os
import glob
import numpy as np


class MakeSomeStupidNoiseMatrices:

    def __call__(self, nr_items: int = 10, folder="./tmp"):
        item_list = list()
        for i in range(nr_items):
            filename = os.path.join(folder, 'test_{}.npy'.format(i))
            with open(filename, 'wb') as f:
                np.save(f, np.random.random(size=(10, )))
                item_list.append(filename)
        return item_list


class StupidAdditionTransformer(BaseEstimator):

    def __init__(self, voxel_size: int = 2):
        self.voxel_size = voxel_size

    def fit(self, X, y):
        return self

    def transform(self, X):
        # load X
        if isinstance(X, str):
            data = np.load(open(X, 'rb'))
            return data + self.voxel_size
        else:
            data_items = list()
            for item in X:
                data = np.load(open(item, 'rb'))
                data_items.append(data + self.voxel_size)
            data_items = np.stack([i for i in data_items], axis=0)
            return data_items


class CachedPhotonPipelineTestsNeuro(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(CachedPhotonPipelineTestsNeuro, cls).setUpClass()

    def setUp(self) -> None:
        super(CachedPhotonPipelineTestsNeuro, self).setUp()
        m = MakeSomeStupidNoiseMatrices()
        self.X = m(folder=self.tmp_folder_path)
        self.y = np.random.randn(len(self.X))
        self.nr_of_expected_pickles_per_config = len(self.X)

    def test_single_subject_caching(self):

        nb = ParallelBranch("subject_caching_test")
        # increase complexity by adding batching
        nb += PipelineElement.create("ResampleImages", StupidAdditionTransformer(), {},
                                     batch_size=4)

        cache_folder = self.cache_folder_path
        cache_folder = os.path.join(cache_folder, 'subject_caching_test')
        nb.base_element.cache_folder = cache_folder

        def transform_and_check_folder(config, expected_nr_of_files):
            nb.set_params(**config)
            nb.transform(self.X, self.y)
            nr_of_generated_cache_files = len(glob.glob(os.path.join(cache_folder, "*.p")))
            self.assertTrue(nr_of_generated_cache_files == expected_nr_of_files)

        # fit with first config
        # expect one cache file per input file
        transform_and_check_folder({'ResampleImages__voxel_size': 5}, self.nr_of_expected_pickles_per_config)

        # after fitting with second config, we expect two times the number of input files to be in cache
        transform_and_check_folder({'ResampleImages__voxel_size': 10}, 2 * self.nr_of_expected_pickles_per_config)

        # fit with first config again, we expect to not have generate other cache files, because they exist
        transform_and_check_folder({'ResampleImages__voxel_size': 5}, 2 * self.nr_of_expected_pickles_per_config)

        # clean up afterwards
        CacheManager.clear_cache_files(cache_folder)
        CacheManager.clear_cache_files(self.tmp_folder_path, force_all=True)

    def test_combi_from_single_and_group_caching(self):

        # 2. specify cache directories
        cache_folder_base = self.cache_folder_path
        cache_folder_neuro = os.path.join(cache_folder_base, 'subject_caching_test')

        CacheManager.clear_cache_files(cache_folder_base)
        CacheManager.clear_cache_files(cache_folder_neuro)

        # 3. set up Neuro Branch
        nb = ParallelBranch("SubjectCaching", nr_of_processes=3)
        # increase complexity by adding batching
        nb += PipelineElement.create("ResampleImages", StupidAdditionTransformer(), {},
                                     batch_size=4)
        nb.base_element.cache_folder = cache_folder_neuro

        # 4. setup usual pipeline
        ss = PipelineElement("StandardScaler", {})
        pca = PipelineElement("PCA", {'n_components': [3, 10, 50]})
        svm = PipelineElement("SVR", {'kernel': ['rbf', 'linear']})

        pipe = PhotonPipeline([('NeuroBranch', nb),
                               ('StandardScaler', ss),
                               ('PCA', pca),
                               ('SVR', svm)])

        pipe.caching = True
        pipe.fold_id = "12345643463434"
        pipe.cache_folder = cache_folder_base

        def transform_and_check_folder(config, expected_nr_of_files_group, expected_nr_subject):
            pipe.set_params(**config)
            pipe.fit(self.X, self.y)
            nr_of_generated_cache_files = len(glob.glob(os.path.join(cache_folder_base, "*.p")))
            self.assertTrue(nr_of_generated_cache_files == expected_nr_of_files_group)

            nr_of_generated_cache_files_subject = len(glob.glob(os.path.join(cache_folder_neuro, "*.p")))
            self.assertTrue(nr_of_generated_cache_files_subject == expected_nr_subject)

        config1 = {'NeuroBranch__ResampleImages__voxel_size': 5, 'PCA__n_components': 7, 'SVR__C': 2}
        config2 = {'NeuroBranch__ResampleImages__voxel_size': 3, 'PCA__n_components': 4, 'SVR__C': 5}

        # first config we expect to have a cached_file for the standard scaler and the pca
        # and we expect to have two files (one resampler, one brain mask) for each input data
        transform_and_check_folder(config1, 2,  self.nr_of_expected_pickles_per_config)

        # second config we expect to have two cached_file for the standard scaler (one time for 5 voxel input and one
        # time for 3 voxel input) and two files two for the first and second config pcas,
        # and we expect to have 2 * nr of input data for resampler plus one time masker
        transform_and_check_folder(config2, 4, 2 * self.nr_of_expected_pickles_per_config)

        # when we transform with the first config again, nothing should happen
        transform_and_check_folder(config1, 4, 2 * self.nr_of_expected_pickles_per_config)

        # when we transform with an empty config, a new entry for pca and standard scaler should be generated, as well
        # as a new cache item for each input data from the neuro branch for each itemin the neuro branch
        with self.assertRaises(ValueError):
            transform_and_check_folder({}, 6, 4 * self.nr_of_expected_pickles_per_config)

        CacheManager.clear_cache_files(cache_folder_base)
        CacheManager.clear_cache_files(cache_folder_neuro)


class CachedHyperpipeTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(CachedHyperpipeTests, cls).setUpClass()

    def setUp(self) -> None:
        super(CachedHyperpipeTests, self).setUp()
        m = MakeSomeStupidNoiseMatrices()
        self.X = m(folder=self.tmp_folder_path)
        self.y = np.random.randn(len(self.X))
        self.nr_of_expected_pickles_per_config = len(self.X)

    def test_neuro_hyperpipe_parallelized_batched_caching(self):

        cache_path = self.cache_folder_path

        self.hyperpipe = Hyperpipe('complex_case',
                                   inner_cv=KFold(n_splits=5),
                                   outer_cv=KFold(n_splits=3),
                                   optimizer='grid_search',
                                   cache_folder=cache_path,
                                   metrics=['mean_squared_error'],
                                   best_config_metric='mean_squared_error',
                                   project_folder=self.tmp_folder_path)

        nb = ParallelBranch("SubjectCaching", nr_of_processes=1)
        nb += PipelineElement.create("ResampleImages", StupidAdditionTransformer(), {'voxel_size': [3, 5, 10]},
                                     batch_size=4)
        self.hyperpipe += nb

        self.hyperpipe += PipelineElement("StandardScaler", {})
        self.hyperpipe += PipelineElement("PCA", {'n_components': [3, 4]})
        self.hyperpipe += PipelineElement("SVR", {'kernel': ['rbf', 'linear']})

        self.hyperpipe.fit(self.X, self.y)

        # assert cache is empty again
        nr_of_p_files = len(glob.glob(os.path.join(self.hyperpipe.cache_folder, "*.p")))
        print(nr_of_p_files)
        self.assertTrue(nr_of_p_files == 0)



from sklearn.datasets import load_boston
from sklearn.model_selection import KFold, ShuffleSplit

from photonai.base import Hyperpipe, PipelineElement, Switch
from photonai.optimization import FloatRange, Categorical, IntegerRange

X, y = load_boston(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('smac_example',
                    optimizer='smac',
                    # possible paramters: facade, intensifier_kwargs
                    # and scenario parameters: https://automl.github.io/SMAC3/master/options.html#scenario
                    optimizer_params={"facade": "SMAC4BO",
                                      "wallclock_limit": 60.0*10,  # seconds
                                      "ta_run_limit": 100},  # limit of configurations
                    metrics=['mean_squared_error', 'pearson_correlation'],
                    best_config_metric='mean_squared_error',
                    outer_cv=ShuffleSplit(n_splits=5, test_size=0.2),
                    inner_cv=KFold(n_splits=5),
                    verbosity=1,
                    project_folder='./tmp/')


# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe.add(PipelineElement('StandardScaler'))
# then do feature selection using a PCA, specify which values to try in the hyperparameter search
my_pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 10)}, test_disabled=True)

switch = Switch("Test_Switch")
# engage and optimize SVR
# linspace and logspace is converted to uniform and log-uniform priors in skopt
switch += PipelineElement('SVR', hyperparameters={'C': FloatRange(0, 10, range_type='linspace'),
                                                   'epsilon': FloatRange(0, 0.0001, range_type='linspace'),
                                                   'tol': FloatRange(1e-4, 1e-2, range_type='linspace'),
                                                   'kernel': Categorical(['linear', 'rbf', 'poly'])})

switch += PipelineElement('RandomForestRegressor', hyperparameters={'n_estimators': Categorical([10, 20])})

my_pipe += switch

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



import os
import unittest
from shutil import rmtree

import numpy as np


class PhotonBaseTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        if hasattr(cls, 'file'):
            cls.base_folder = os.path.dirname(os.path.abspath(cls.file))

            cls.cache_folder_path = os.path.join(cls.base_folder, "cache")
            os.makedirs(cls.cache_folder_path, exist_ok=True)

            cls.tmp_folder_path = os.path.join(cls.base_folder, "tmp")
            os.makedirs(cls.tmp_folder_path, exist_ok=True)

            cls.dask_path = os.path.join(cls.base_folder, "dask-worker-space")
            cls.photon_setup_error_path = os.path.join(cls.base_folder, "photon_setup_errors.log")

    def setUp(self) -> None:
        pass

    @classmethod
    def tearDownClass(cls) -> None:
        rmtree(cls.cache_folder_path, ignore_errors=True)
        rmtree(cls.tmp_folder_path, ignore_errors=True)
        rmtree(cls.dask_path, ignore_errors=True)
        if os.path.isfile(cls.photon_setup_error_path):
            os.remove(cls.photon_setup_error_path)


def elements_to_dict(elements):
    if isinstance(elements, dict):
        new_dict = dict()
        for name, element in elements.items():
            new_dict[name] = elements_to_dict(element)
        elements = new_dict
    elif isinstance(elements, list):
        new_list = list()
        for element in elements:
            new_list.append(elements_to_dict(element))
        elements = new_list
    elif isinstance(elements, tuple):
        new_list = list()
        for element in elements:
            new_list.append(elements_to_dict(element))
        elements = tuple(new_list)
    elif hasattr(elements, '__dict__'):
        new_dict = dict()
        elements = elements.__dict__
        for name, element in elements.items():
            new_dict[name] = elements_to_dict(element)
        elements = new_dict
    else:
        if not isinstance(elements, (str, float, int, complex, np.ndarray)):
            return None
    return elements



from sklearn.model_selection import ShuffleSplit
from sklearn.datasets import fetch_openml

from photonai.base import Hyperpipe, PipelineElement, Preprocessing
from photonai.optimization import FloatRange

audiology = fetch_openml(name='audiology')
X = audiology.data.values
y = audiology.target.values

my_pipe = Hyperpipe('hot_encoder_pipeline',
                    inner_cv=ShuffleSplit(n_splits=5, test_size=0.2),
                    outer_cv=ShuffleSplit(n_splits=3, test_size=0.2),
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 20},
                    metrics=['accuracy'],
                    best_config_metric='accuracy',
                    allow_multidim_targets=True,
                    project_folder='./tmp')

pre_proc = Preprocessing()
pre_proc += PipelineElement('OneHotEncoder', sparse=False)
pre_proc += PipelineElement('LabelEncoder')
my_pipe += pre_proc

my_pipe += PipelineElement('PCA', hyperparameters={'n_components': FloatRange(0.2, 0.7)})
my_pipe += PipelineElement('SVC', hyperparameters={'C': FloatRange(1, 150)}, kernel='rbf')

my_pipe.fit(X, y)



import datetime
import logging
import sys
import sklearn

logging.getLogger(sklearn.__name__).setLevel(logging.ERROR)

dask_logger = logging.getLogger('distributed.utils_perf')
dask_logger.setLevel(logging.ERROR)
for handler in dask_logger.handlers:
    handler.setLevel(logging.ERROR)

# create photon logger
logger = logging.getLogger('PHOTONAI')
handler = logging.StreamHandler(stream=sys.stdout)
logger.addHandler(handler)
logger.propagate = False

VERBOSE_LEVELV_NUM = 25
CLEAN_LEVELV_NUM = 21
INFO_LEVELV_NUM = 20
DEBUG_LEVELV_NUM = 10
LENGTH = 101


# add custom log level
def photon_system_log(self, message, *args, **kws):
    if self.isEnabledFor(VERBOSE_LEVELV_NUM):
        self._log(VERBOSE_LEVELV_NUM, message, args, **kws)


def star_log(self):
    photon_system_log(self, '*' * LENGTH)


def system_line(self):
    photon_system_log(self, '-' * LENGTH)


def underscore_log(self):
    clean_info(self, '-' * LENGTH)


def clean_info(self, message, *args, **kws):
    if self.isEnabledFor(CLEAN_LEVELV_NUM):
        self._log(CLEAN_LEVELV_NUM, message, args, **kws)


def info(self, message, *args, **kws):
    if self.isEnabledFor(INFO_LEVELV_NUM):
        timestamp = datetime.datetime.now()
        log_message = timestamp.strftime("%d/%m/%Y-%H:%M:%S")
        if message:
            log_message += " | " + str(message)
        self._log(INFO_LEVELV_NUM, log_message, args, **kws)


def debug(self, message, *args, **kws):
    if self.isEnabledFor(DEBUG_LEVELV_NUM):
        timestamp = datetime.datetime.now()
        if isinstance(message, str):
            log_message = timestamp.strftime("%d/%m/%Y-%H:%M:%S")
            log_message += " | " + str(message)
        else:
            log_message = message
        self._log(DEBUG_LEVELV_NUM, log_message, args, **kws)


logging.addLevelName(VERBOSE_LEVELV_NUM, "PHOTON_SYSTEM_LOG")
logging.Logger.photon_system_log = photon_system_log
logging.addLevelName(CLEAN_LEVELV_NUM, "CLEAN_INFO")
logging.Logger.clean_info = clean_info
logging.addLevelName(INFO_LEVELV_NUM, "INFO")
logging.Logger.info = info
logging.addLevelName(DEBUG_LEVELV_NUM, "DEBUG")
logging.Logger.debug = debug
logging.Logger.stars = star_log
logging.Logger.line = underscore_log
logging.Logger.system_line = system_line



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from photonai.photonlogger.logger import logger

try:
    from imblearn import over_sampling, under_sampling, combine
    __found__ = True
except ModuleNotFoundError:
    __found__ = False


class ImbalancedDataTransformer(BaseEstimator, TransformerMixin):
    """
    Applies the chosen strategy to the data in order to balance the input data.
    Instantiates the strategy filter object according to the name given as string literal.
    Underlying architecture: Imbalanced-Learning.
    More information on their [documentation](https://imbalanced-learn.org/stable/).

    Example:
        ``` python
        from photonai.optimization import Categorical

        tested_methods = Categorical(['RandomOverSampler', 'SMOTEENN', 'SVMSMOTE',
                              'BorderlineSMOTE', 'SMOTE', 'ClusterCentroids'])
        PipelineElement('ImbalancedDataTransformer',
                        hyperparameters={'method_name': tested_methods},
                        test_disabled=True)
        ```

    """
    IMBALANCED_DICT = {
        'oversampling': ["ADASYN",
                         "BorderlineSMOTE",
                         "KMeansSMOTE",
                         "RandomOverSampler",
                         "SMOTE",
                         "SMOTENC",
                         "SVMSMOTE"],
        'undersampling': ["AllKNN",
                          "ClusterCentroids",
                          "CondensedNearestNeighbour",
                          "EditedNearestNeighbours",
                          "InstanceHardnessThreshold",
                          "NearMiss",
                          "NeighbourhoodCleaningRule",
                          "OneSidedSelection",
                          "TomekLinks",
                          "RandomUnderSampler",
                          "RepeatedEditedNearestNeighbours"],
        'combine': ["SMOTEENN", "SMOTETomek"],
    }

    def __init__(self, method_name: str = 'RandomUnderSampler', **kwargs):
        """
        Instantiates an object that transforms the data into balanced groups according to the given method.

        Parameters:
            method_name:
                Imbalanced learning strategy. Possible values with

                - an oversampling strategy are:
                    - ADASYN,
                    - BorderlineSMOTE,
                    - KMeansSMOTE,
                    - RandomOverSampler,
                    - SMOTE,
                    - SMOTENC,
                    - SVMSMOTE,

                - an undersampling strategy are:
                    - ClusterCentroids,
                    - RandomUnderSampler,
                    - NearMiss,
                    - InstanceHardnessThreshold,
                    - CondensedNearestNeighbour,
                    - EditedNearestNeighbours,
                    - RepeatedEditedNearestNeighbours,
                    - AllKNN,
                    - NeighbourhoodCleaningRule,
                    - OneSidedSelection,

                - a combined strategy are:
                    - SMOTEENN,
                    - SMOTETomek.

            **kwargs:
                Any parameters to pass to the imbalance strategy object.

        """
        if not __found__:
            raise ModuleNotFoundError("Module imblearn not found or not installed as expected. "
                                      "Please install the requirements.txt in PHOTON main folder.")

        self.method_name = method_name
        self.needs_y = True

        imbalance_type = ''
        for group, possible_strategies in ImbalancedDataTransformer.IMBALANCED_DICT.items():
            if self.method_name in possible_strategies:
                imbalance_type = group

        if imbalance_type == "oversampling":
            home = over_sampling
        elif imbalance_type == "undersampling":
            home = under_sampling
        elif imbalance_type == "combine" or imbalance_type == "combination":
            home = combine
        else:
            msg = "Imbalance Type not found. Can be oversampling, undersampling or combine. " \
                  "Oversampling: method_name one of {}. Undersampling: method_name one of {}." \
                  "Combine: method_name one of {}.".format(str(self.IMBALANCED_DICT["oversampling"]),
                                                           str(self.IMBALANCED_DICT["undersampling"]),
                                                           str(self.IMBALANCED_DICT["combine"]))
            logger.error(msg)
            raise ValueError(msg)

        desired_class = getattr(home, method_name)
        self.method = desired_class(**kwargs)

    def fit_transform(self, X: np.ndarray, y: np.ndarray = None, **kwargs) -> (np.ndarray, np.ndarray):
        """
        Call of the underlying imblearn.fit_resample(X, y).

        Parameters:
            X:
                The input samples of shape [n_samples, n_features].

            y:
                The input targets of shape [n_samples, 1].

            **kwargs:
                Ignored input.

        Returns:
            Transformed data.

        """
        return self.method.fit_resample(X, y)

    #  define an alias for imblearn consistency
    fit_sample = fit_transform
    fit_resample = fit_transform

    def fit(self, X, y, **kwargs):
        """Empty method required in PHOTONAI."""
        return

    def transform(self, X: np.ndarray, y: np.ndarray = None, **kwargs) -> (np.ndarray, np.ndarray):
        """
        Forwarding to the self.fit_transform method.

        Parameters:
            X:
                The input samples of shape [n_samples, n_features].

            y:
                The input targets of shape [n_samples, 1].

            **kwargs:
                Ignored input.

        Returns:
            Transformed data.

        """
        return self.fit_transform(X, y)
